from live.constants import *

import live.query
import live.object

import random

class Clip(live.LoggingObject):
	""" An object representing a single clip in a Live set.

	Properties:
	track -- Track object this clip resides within.
	index -- index of this clip
	length -- length of cip, in beats
	name -- human-readable name
	state -- one of CLIP_STATE_EMPTY, CLIP_STATE_STOPPED, CLIP_STATE_PLAYING
	"""

	def __init__(self, track, index, length = 4):
		""" Create a new clip.
		
		Arguments:
		track -- a Track or Group object
		index -- index of this clip within the track
		length -- length of clip, in beats
		"""
		self.track = track
		self.index = index
		self.length = length
		self.looplen = length
		self.indent = 2
		self.state = CLIP_STATUS_STOPPED
		self.name = None

	def __str__(self):
		name = ": %s" % self.name if self.name else ""
		state_symbols = { CLIP_STATUS_EMPTY : " ", CLIP_STATUS_STOPPED : "-", CLIP_STATUS_PLAYING : ">", CLIP_STATUS_STARTING : "*" }
		state_symbol = state_symbols[self.state]
		
		return "live.clip(%d,%d)%s [%s]" % (self.track.index, self.index, name, state_symbol)

	def dump(self):
		""" Output a human-readable description of this clip. """
		self.trace("(len = %d, looplen = %d)" % (self.length, self.looplen))

	@property
	def set(self):
		""" Helper method to obtain the Set that this clip resides within. """
		return self.track.set

	def _syncopate(self):
		looplen_old = self.looplen
		if self.looplen == self.length:
			self.looplen -= 1
		elif self.looplen == 1:
			self.looplen += 1
		else:
			self.looplen += random.choice([ -1, 1 ])
		self.trace("syncopating loop length from %d to %d (total length %d)" % (looplen_old, self.looplen, self.length))
		self.set.set_clip_loop_end(self.track.index, self.index, self.looplen)

	def reset(self):
		if self.looplen != self.length:
			self.trace("resetting loop length to %d" % self.length)
			self.set.set_clip_loop_end(self.track.index, self.index, self.looplen)

	def play(self):
		""" Start playing clip. """
		self.trace("playing")
		self.set.play_clip(self.track.index, self.index)
		self.track.playing = True
		if type(self.track) is live.Group:
			for track in self.track.tracks:
				#------------------------------------------------------------------------
				# when we trigger a group clip, it triggers each of the corresponding
				# clips in the tracks it contains. thus need to update the playing
				# status of each of our tracks, assuming that all clips/stop buttons
				# are enabled.
				#------------------------------------------------------------------------
				if self.index < len(track.clips) and track.clips[self.index] is not None:
					track.playing = True
				else:
					track.playing = False

	def stop(self):
		""" Stop playing clip """
		self.trace("stopping")
		self.set.stop_clip(self.track.index, self.index)
		self.track.playing = False
	
	def get_pitch(self):
		return self.set.get_clip_pitch(self.track.index, self.index)
	def set_pitch(self, coarse, fine = 0):
		self.set.set_clip_pitch(self.track.index, self.index, coarse, fine)

	pitch = property(get_pitch, set_pitch, doc = "Pitch (coarse, fine)")

	def get_muted(self):
		return bool(self.set.get_clip_mute(self.track.index, self.index))
	def set_muted(self, muted = True):
		self.set.set_clip_mute(self.track.index, self.index, int(muted))

	muted = property(get_muted, set_muted, doc = "Muted (boolean)")

	def get_next_clip(self, wrap = False, allow_gaps = True):
		""" Return the next clip in the track, or None if not found.
		wrap -- If set, will wrap from (N-1) to 0 and vice versa.
		allow_gaps -- If unset, requires that clip slots must be contiguous.
		"""
		clips = self.track.clips
		index = clips.index(self)
		clip_range = []
		for n in range(index, len(clips)):
			if clips[n] is None:
				if not allow_gaps:
					break
			else:
				clip_range.append(clips[n])
		for n in reversed(range(0, index)):
			if clips[n] is None:
				if not allow_gaps:
					break
			else:
				clip_range.insert(0, clips[n])

		index = clip_range.index(self)
		next_index = index + 1
		if wrap:
			next_index = next_index % len(clip_range)
		if next_index < len(clip_range):
			return clip_range[next_index]
		return None

	def get_prev_clip(self, wrap = False, allow_gaps = True):
		""" Return the previous clip in the track, or None if not found.
		wrap -- If set, will wrap from (N-1) to 0 and vice versa.
		allow_gaps -- If unset, requires that clip slots must be contiguous.
		"""
		clips = self.track.clips
		index = clips.index(self)
		clip_range = []
		for n in range(index, len(clips)):
			if clips[n] is None:
				if not allow_gaps:
					break
			else:
				clip_range.append(clips[n])
		for n in reversed(range(0, index)):
			if clips[n] is None:
				if not allow_gaps:
					break
			else:
				clip_range.insert(0, clips[n])

		index = clip_range.index(self)
		prev_index = index - 1
		if wrap and prev_index < 0:
			prev_index += len(clip_range)
		if prev_index >= 0:
			return clip_range[prev_index]
		return None



from live.object import *

class Group (LoggingObject):
	""" Represents a grouped set of Track objects.
	Because we can't programmatically query whether a track in Live is an
	individual or group track, the user must name groups using a special
	format which is passed to set.scan(group_re = RE)

	Properties:
	track_index -- The numerical track index of this group
	group_index -- Groups are auto-numbered from 0
	name -- Human-readable name
	tracks -- List of Track objects contained within this group
	"""

	def __init__(self, set, track_index, group_index, name):
		self.set = set
		self.track_index = track_index
		self.group_index = group_index

		# needed so that Clip objects can call the 'index' method on Group and Track accordingly
		# TODO: rename 'index' to 'track_index' on Track objects too
		self.index = track_index
		self.indent = 1
		self.name = name
		self.group = None
		self.tracks = []
		self.clips = []

	def __str__(self):
		string = "live.group(%d): %s" % (self.group_index, self.name)
		if len(self.tracks):
			string = string + " [tracks %d-%d]" % (self.tracks[0].index, self.tracks[len(self.tracks) - 1].index)
		return string

	@property
	def scene_indexes(self):
		indexes = {}
		for track in self.tracks:
			for index in track.scene_indexes:
				indexes[index] = 1
		indexes = indexes.keys()
		indexes = sorted(indexes)
		return indexes

	def add_track(self, track):
		""" Append a new track to this group. Should probably only be called by Set.scan(). """
		self.tracks.append(track)

	def dump(self):
		self.trace("%d tracks" % len(self.tracks))
		for track in self.tracks:
			track.dump()

	def play_clip(self, clip_index):
		""" Start playing group clip. """
		self.clips[clip_index].play()

	def stop(self):
		""" Immediately stop group from playing. """
		self.playing = False
		self.set.stop_track(self.track_index)

	@property
	def active_clips(self):
		""" Return a dictionary of all non-empty clipslots: { index : Clip, ... } """
		active_clips = filter(lambda n: n is not None, self.clips)
		return active_clips
	get_active_clips = active_clips




from live.constants import *
from live.object import *
from live.clip import *
from live.query import *

import re
import random

class Track(LoggingObject):
	""" Represents a single Track, either audio or MIDI.
	Resides within a Set, and contains one or more Device and Clip objects.
	May be contained within a Group.

	Properties:
	set -- The containing Set object
	index -- The numerical index of this Track within the Set
	name -- Human-readable name
	group -- (Optional) reference to containing Group object
	clips -- Dictionary of contained Clips: { index : Clip }
	devices -- List of contained Devices
	"""

	def __init__(self, set, index, name, group = None):
		self.set = set
		self.index = index
		self.name = name
		self.group = group
		self.indent = 2 if self.group else 1

		self.clip_init = None
		self.clip_playing = None
		self.clips = []
		self.devices = []
		self.playing = False

	def __str__(self):
		if self.group:
			return "live.track(%d,%d): %s" % (self.group.group_index, self.index, self.name)
		else:
			return "live.track(%d): %s" % (self.index, self.name)

	def dump(self):
		self.trace()
		if self.devices:
			for device in self.devices:
				device.dump()
		active_clips = self.active_clips
		for clip in active_clips:
			clip.dump()

	@property
	def scene_indexes(self):
		""" TODO: turn this into something that returns Scene objects (which don't yet exist) """
		indexes = []
		for clip in self.active_clips:
			indexes.append(clip.index)
		return indexes

	def clips_between(self, index_start, index_finish):
		""" Returns a list of Clip objects between index_start and index_finish """
		clips = []
		for n in range(index_start, index_finish):
			if n in self.clips:
				clips.append(self.clips[n])
		return clips

	def play_clip(self, index):
		""" Plays clip of given index. """
		self.playing = True
		self.clip_playing = index
		clip = self.clips[index]
		clip.play()

	def play_clip_random(self):
		""" Plays a random clip. """
		if len(self.clips.keys()) == 0:
			self.warn("no clips found on track, returning")
			return
		index = random.choice(self.clips.keys())
		self.play_clip(index)

	def stop(self):
		""" Immediately stop track from playing. """
		self.playing = False
		self.clip_playing = None
		self.set.stop_track(self.index)

	def syncopate(self):
		if self.clip_playing is not None:
			self.clips[self.clip_playing].syncopate()
		else:
			self.warn("asked to syncopate but not yet playing!")

	def has_clip(self, index):
		""" Determine whether this track contains a clip at slot index. """
		return self.clips.has_key(index)

	def get_clips(self):
		return self.clips

	@property
	def active_clips(self):
		""" Return a dictionary of all non-empty clipslots: { index : Clip, ... } """
		active_clips = filter(lambda n: n is not None, self.clips)
		return active_clips
	get_active_clips = active_clips

	def clip_playing(self):
		""" Return the currently playing Clip, or None. """
		for clip in self.clips:
			if clip.state == CLIP_STATE_PLAYING:
				return clip
		return None

	def sync(self):
		# XXX: why is this needed?
		info = self.set.get_track_info()
		print "track %d: info %s" % (self.index, info[0])
		self.playing = True if info[0] > 1 else False

	def walk(self):
		""" Move forward or backwards between clips. """
		if not self.playing:
			if self.clip_init:
				self.trace("walking to initial clip %d" % self.clip_init)
				self.play_clip(self.clip_init)
			else:
				self.warn("no clips found on track %d, returning" % self.index)
				return
		else:
			options = []
			if self.clips.has_key(self.clip_playing - 1):
				options.append(self.clip_playing - 1)
			if self.clips.has_key(self.clip_playing + 1):
				options.append(self.clip_playing + 1)

			if len(options) > 0:
				index = random.choice(options)
				self.trace("walking from clip %d to %d" % (self.clip_playing, index))
				self.play_clip(index)
			else:
				self.trace("walking to random clip")
				self.play_clip_random()

	def scan_clip_names(self):
		#--------------------------------------------------------------------------
		# scan for clip names.
		# is nice, but slows things down significantly -- so disable by default.
		#--------------------------------------------------------------------------
		for clip in self.active_clips:
			clip_name = self.set.get_clip_name(self.index, clip.index)
			clip.name = clip_name
			self.trace("scan_clip_names: (%d, %d) -> %s" % (self.index, clip.index, clip.name))

	#------------------------------------------------------------------------
	# get/set: volume
	#------------------------------------------------------------------------

	def set_volume(self, value):
		self.set.set_track_volume(self.index, value)
	def get_volume(self):
		return self.set.get_track_volume(self.index)
	volume = property(get_volume, set_volume, None, "track volume (0..1)")

	#------------------------------------------------------------------------
	# get/set: pan
	#------------------------------------------------------------------------

	def set_pan(self, value):
		self.set.set_track_pan(self.index, value)
	def get_pan(self):
		return self.set.get_track_pan(self.index)
	pan = property(get_pan, set_pan, None, "track pan (-1..1)")

	#------------------------------------------------------------------------
	# get/set: mute
	#------------------------------------------------------------------------

	def set_mute(self, value):
		self.set.set_track_mute(self.index, value)
	def get_mute(self):
		return self.set.get_track_mute(self.index)
	mute = property(get_mute, set_mute, None, "track mute (0/1)")

	#------------------------------------------------------------------------
	# get/set: arm
	#------------------------------------------------------------------------

	def set_arm(self, value):
		self.set.set_track_arm(self.index, value)
	def get_arm(self):
		return self.set.get_track_arm(self.index)
	arm = property(get_arm, set_arm, None, "track armed to record (0.1)")

	#------------------------------------------------------------------------
	# get/set: solo
	#------------------------------------------------------------------------

	def set_solo(self, value):
		self.set.set_track_solo(self.index, value)
	def get_solo(self):
		return self.set.get_track_solo(self.index)
	solo = property(get_solo, set_solo, None, "track in solo mode (0/1)")

	#------------------------------------------------------------------------
	# get/set: send
	#------------------------------------------------------------------------

	def set_send(self, send_index, value):
		""" Set the send level of the given send_index (0..1) """
		self.set.set_track_send(self.index, send_index, value)
	def get_send(self, send_index):
		""" Get the send level of the given send_index (0..1) """
		return self.set.get_track_send(self.index, send_index)




""" Unit tests for PyLive """

import unittest
import random
import live
import math
import time

class Tests(unittest.TestCase):
	def testSetTempo(self):
		""" set and query tempo """
		set = live.Set()
		tempo = random.randint(50, 150)
		set.tempo = tempo
		self.assertEqual(tempo, set.tempo)

	def testSetTime(self):
		""" set and query arrangement time """
		set = live.Set()
		t = random.randint(0, 16)
		set.set_time(t)
		time.sleep(0.2)
		t_new = set.get_time()
		self.assertEqual(t, t_new)


if __name__ == "__main__":
	unittest.main()



import live.query
import live.object

import random

class Device(live.LoggingObject):
	""" Represents an instrument or audio effect residing within a Track.
	Contains one one or more Parameters.

	Properties:
	track -- Track object that this Device resides within
	index -- Numeric index of this device
	name -- Human-readable name
	parameters -- List of Parameter objects
	"""

	def __init__(self, track, index, name):
		self.track = track
		self.index = index
		self.name = name
		self.parameters = []
		self.indent = 2

	def __str__(self):
		return "live.device(%d,%d): %s (%d parameters)" % (self.track.index, self.index, self.name, len(self.parameters))

	def dump(self):
		self.trace()

	@property
	def set(self):
		""" Helper function to return the Set that this Device resides within. """
		return self.track.set

	def set_parameter(self, index, value):
		if type(index) == int:
			parameter = self.parameters[index]
		else:
			parameter = next(p for p in self.parameters if p.name == index)
		parameter.value = value

	def get_parameter(self, index):
		if type(index) == int:
			parameter = self.parameters[index]
		else:
			parameter = next(p for p in self.parameters if p.name == index)
		return parameter.value



import live.query
import live.object

import random

class Parameter(live.LoggingObject):
	""" Represents a parameter of a Live device (either an instrument or
	effects unit.

	Properties:
	device -- the Device object that this Parameter belongs to
	index -- numerical index
	name -- name, as specified by the device 
	value -- current value, within a parameter-specific range

	minimum -- minimum value (float or int)
	maximum -- maximum value (float or int)
	"""

	def __init__(self, device, index, name, value):
		self.device = device
		self.index = index
		self.name = name
		self.value = value
		self.minimum = 0.0
		self.maximum = 1.0
		self.indent = 3

	def __str__(self):
		return "live.parameter(%d,%d,%d): %s (range %.3f-%.3f)" % (self.device.track.index, self.device.index, self.index, self.name, self.minimum, self.maximum)

	def is_integer(self):
		# TODO: fix for enums (such as Quality in Reverb unit). how?
		return self.name.endswith("On")

	@property
	def set(self):
		""" Helper function to return the Set that this parameter resides within. """
		return self.device.track.set

	@property
	def track(self):
		""" Helper function to return the Track that this parameter resides within. """
		return self.device.track

	def dump(self):
		self.trace()

	def set_value(self, value):
		self.set.set_device_param(self.device.track.index, self.device.index, self.index, value)

	def get_value(self):
		return self.device.track.set.get_device_param(self.device.track.index, self.device.index, self.index)

	value = property(get_value, set_value)

	def randomise(self):
		""" Set the parameter's value to a uniformly random value within [minimum,
		maximum] """

		if self.is_integer():
			value = random.randint(self.minimum, self.maximum)
		else:
			value = random.uniform(self.minimum, self.maximum)
		self.value = value



"""

pylive
~~~~~~

A framework for controlling Ableton Live from a standalone Python script,
mediated via Open Sound Control.

USAGE
~~~~~

import live

set = live.Set()
set.scan(scan_clip_names = True)
set.tempo = 110.0

track = set.tracks[0]
print "track name %s" % track.name

clip = track.clips[0]
print "clip name %s, length %d beats" % (clip.name, clip.length)
clip.play()

"""

__version__ = "0.1.1"
__author__ = "Daniel Jones <http://www.erase.net/>"
__all__ = [ "Query", "Set", "Track", "Group", "Clip", "Device", "Parameter" ]

debug = False

from object import *
from constants import *
from set import *
from query import *
from track import *
from group import *
from clip import *
from scene import *
from device import *
from parameter import *



from live.constants import *

import live.query
import live.object

import random

class Scene(live.LoggingObject):
	""" An object representing a single scene in a Live set.

	Properties:
	index -- index of this scene
	name -- human-readable name
	"""

	def __init__(self, set, index):
		""" Create a new scene.
		
		Arguments:
		index -- index of this scene
		"""
		self.set = set
		self.index = index
		self.name = None
		self.indent = 1

	def __str__(self):
		name = ": %s" % self.name if self.name else ""
		
		return "live.scene(%d)%s" % (self.index, name)

	def dump(self):
		self.trace()

	def play(self):
		""" Start playing scene. """
		self.trace("playing")
		self.set.play_scene(self.index)



import time
import signal
import inspect
import threading

from OSC import *
from live.object import *

class LiveError(Exception):
	"""Base class for all Live-related errors
	"""
	def __init__(self, message):
		self.message = message

	def __str__(self):
		return self.message


def singleton(cls):
	instances = {}
	def getinstance(*args):
		if cls not in instances:
			instances[cls] = cls(*args)
		return instances[cls]
	return getinstance

#------------------------------------------------------------------------
# Helper methods to save instantiating an object when making calls.
#------------------------------------------------------------------------

def query(*args, **kwargs):
	return Query().query(*args, **kwargs)

def query_one(*args, **kwargs):
	return Query().query_one(*args, **kwargs)

def cmd (*args, **kwargs):
	Query().cmd(*args, **kwargs)

@singleton
class Query(LoggingObject):
	""" Object responsible for passing OSC queries to the LiveOSC server,
	parsing and proxying responses.

	This object is a singleton, under the assumption that only one Live instance
	can be running, so only one global Live Query object should be needed.

	Following this assumption, static helper functions also exist:

		live.query(path, *args)
		live.query_one(path, *args)
		live.cmd(path, *args)
	"""

	def __init__(self, address = ("localhost", 9000), listen_port = 9001):
		self.indent = 0
		self.beat_callback = None
		self.startup_callback = None
		self.listening = False
		self.listen_port = listen_port

		# handler callbacks for particular messages from Live.
		# used so that other processes can register callbacks when states change.
		self.handlers = {}

		self.osc_address = address
		self.osc_client = OSCClient()
		self.osc_client.connect(address)
		try:
			self.osc_server = OSCServer(("localhost", self.listen_port))
			self.osc_server.print_tracebacks = True
			self.osc_server_thread = None
		except:
			"Couldn't connect to Live (is another process already connected?)"
		self.osc_read_event = None
		self.osc_timeout = 5

		self.response_address = None

		self.listen()

	def __str__(self):
		return "live.query"

	def stop(self):
		""" Terminate this query object and unbind from OSC listening. """
		if self.listening:
			self.osc_server.close()
			self.listening = False

	def listen(self):
		""" Commence listening for OSC messages from LiveOSC. """
		if self.listening:
			return

		try:
			self.trace("started listening")
			self.osc_server.addMsgHandler("default", self.handler)
			self.osc_server_thread = threading.Thread(target = self.osc_server.serve_forever)
			self.osc_server_thread.setDaemon(True)
			self.osc_server_thread.start()
			self.listening = True
		except Exception, e:
			self.warn("listen failed (couldn't bind to port %d): %s" % (self.listen_port, e))

	def cmd(self, msg, *args):
		""" Send a Live command without expecting a response back:

			live.cmd("/live/tempo", 110.0) """
		
		msg = OSCMessage(msg)
		msg.extend(list(args))
		self.debug("OSC: %s %s", msg, args)
		try:
			self.osc_client.send(msg)
		except OSCClientError:
			raise LiveError("Couldn't send message to Live (is LiveOSC present and activated?)")

	def query(self, msg, *args, **kwargs):
		""" Send a Live command and synchronously wait for its response:

			return live.query("/live/tempo")

		Returns a list of values. """

		#------------------------------------------------------------------------
		# use **kwargs because we want to be able to specify an optional kw
		# arg after variable-length args -- 
		# eg live.query("/set/freq", 440, 1.0, response_address = "/verify/freq")
		# http://stackoverflow.com/questions/5940180/python-default-keyword-arguments-after-variable-length-positional-arguments
		#------------------------------------------------------------------------
		if not self.listening:
			self.listen()

		#------------------------------------------------------------------------
		# some calls produce responses at different addresses
		# (eg /live/device -> /live/deviceall). specify a response_address to
		# take account of this.
		#------------------------------------------------------------------------
		response_address = kwargs.get("response_address", None)
		if response_address:
			self.response_address = response_address
		else:
			self.response_address = msg

		self.query_rv = []

		self.osc_server_event = threading.Event()

		msg = OSCMessage(msg)
		msg.extend(list(args))
		self.debug("OSC: %s %s", msg, args)
		try:
			self.osc_client.send(msg)
		except OSCClientError:
			raise LiveError("Couldn't send message to Live (is LiveOSC present and activated?)")

		rv = self.osc_server_event.wait(self.osc_timeout)
		if not rv:
			print "*** timed out waiting for server response"

		return self.query_rv

	def query_one(self, msg, *args):
		""" Send a Live command and synchronously wait for its response:

			return live.query_one("/live/tempo")

		Returns a single scalar value (in this case, a float). """
		rv = self.query(msg, *args)
		if not rv:
			return None
		return rv[0]

	def handler(self, address, tags, data, source):
		self.debug("OSC: %s %s" % (address, data))
		#------------------------------------------------------------------------
		# Execute any callbacks that have been registered for this message
		#------------------------------------------------------------------------
		if address in self.handlers:
			for handler in self.handlers[address]:
				handler(*data)

		#------------------------------------------------------------------------
		# If this message is awaiting a synchronous return, trigger the
		# thread event and update our return value. 
		#------------------------------------------------------------------------
		if address == self.response_address:
			self.query_rv += data
			self.osc_server_event.set()
			return

		if address == "/live/beat":
			if self.beat_callback is not None:
				#------------------------------------------------------------------------
				# Beat callbacks are used if we want to trigger an event on each beat,
				# to synchronise with the timing of the Live set.
				#
				# Callbacks may take one argument: the current beat count.
				# If not specified, call with 0 arguments.
				#------------------------------------------------------------------------
				# It might be nice to send the current beat # as a parameter, but we
				# also want to be able to handle callbacks with no args -- TODO: look
				# into this.
				#------------------------------------------------------------------------
				# argspec = inspect.getargspec(self.beat_callback)
				# if len(argspec.args) > 0:
				#	self.beat_callback(data[0])
				#------------------------------------------------------------------------
				self.beat_callback()
		elif address == "/remix/oscserver/startup":
			if self.startup_callback is not None:
				self.startup_callback()

	def add_handler(self, address, handler):
		if not address in self.handlers:
			self.handlers[address] = []
		self.handlers[address].append(handler)




#!/usr/bin/env python

from setuptools import setup

#------------------------------------------------------------------------
# Generate a PyPI-friendly RST file from our markdown README.
#------------------------------------------------------------------------
try:
	import pypandoc
	long_description = pypandoc.convert('README.md', 'rst')
except:
	long_description = None

setup(
	name = 'pylive',
	version = '0.1.4',
	description = 'Python remote control of Ableton Live',
	long_description = long_description,
	author = 'Daniel Jones',
	author_email = 'dan-pylive@erase.net',
	url = 'https://github.com/ideoforms/pylive',
	packages = ['live'],
	install_requires = ['pyliblo >= 0.9.1'],
	keywords = ('sound', 'music', 'ableton', 'osc'),
	classifiers = [
		'Topic :: Multimedia :: Sound/Audio',
		'Topic :: Artistic Software',
		'Topic :: Communications',
		'Development Status :: 4 - Beta',
		'Intended Audience :: Developers'
	]
)
